#ifndef MIE_ETAT_IF_H_
#define MIE_ETAT_IF_H_
/**
	@file etat_if.h
	@brief etat pairing interface
	@author herumi
	@version $Revision: 1.15 $
	@url http://homepage1.nifty.com/herumi/crypt/pairing.html
	@date $Date: 2010/03/08 02:16:19 $
	@note modified new BSD license
	http://www.opensource.org/licenses/bsd-license.php
*/
#include <stdlib.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef enum {
	MIE_NO_ERR = 0,
	MIE_NOT_SUPPORT,
	MIE_NO_MEMORY,
	MIE_NO_SSSE3,
	MIE_SMALL_SIZE,
	MIE_END_OF_ERR
} MIE_Err;

typedef enum {
	MIE_ETAT_193,
	MIE_END_OF_PAIRING_PARAMETER
} MIE_PairingParameter;
/**
	convert error to string
	@param err [in] err number
*/
const char *MIE_ErrToStr(MIE_Err err);

/**
	Fp := F_3^m
*/
typedef struct MIE_Fp MIE_Fp;

/**
	Fq := (F_3^m)^6
*/
typedef struct MIE_Fq MIE_Fq;

/**
	Ec : y^2 = x^3 - x + b on Fp
	(b = 1)
*/
typedef struct MIE_Ec MIE_Ec;

/**
	functions for pairing
*/
typedef struct MIE_System {
	// destructor of system
	void (*destroySystem)(struct MIE_System*);

	// functions for Fp
	MIE_Fp *(*FpCreate)(void);
	MIE_Fp *(*FpCreateByStr)(const char *str);
	void (*FpDestroy)(MIE_Fp *in);

	MIE_Err (*FpToStr)(char *buf, size_t bufSize, const MIE_Fp *in);

	int (*FpEqual)(const MIE_Fp *in1, const MIE_Fp *in2);
	void (*FpClear)(MIE_Fp *in);

	void (*FpSet)(MIE_Fp *out, const MIE_Fp *in);
	void (*FpSetByStr)(MIE_Fp *out, const char *str);

	void (*FpAdd)(MIE_Fp *out, const MIE_Fp *in1, const MIE_Fp *in2);
	void (*FpSub)(MIE_Fp *out, const MIE_Fp *in1, const MIE_Fp *in2);
	void (*FpMul)(MIE_Fp *out, const MIE_Fp *in1, const MIE_Fp *in2);

	void (*FpInv)(MIE_Fp *out, const MIE_Fp *in);
	void (*FpNeg)(MIE_Fp *out, const MIE_Fp *in);
	/**
		calc square root of in
		@retval 1 find
		@retval 0 not exist
	*/
	int (*FpSqrt)(MIE_Fp *out, const MIE_Fp *in);

	void (*FpPow)(MIE_Fp *out, const MIE_Fp *in, const unsigned int *tbl, size_t tblSize);


	// functions for Fq
	MIE_Fq *(*FqCreate)(void);
	void (*FqDestroy)(MIE_Fq *in);

	MIE_Err (*FqToStr)(char *buf, size_t bufSize, const MIE_Fq *in);
	int (*FqEqual)(const MIE_Fq *in1, const MIE_Fq *in2);
	void (*FqClear)(MIE_Fq *in);

	void (*FqSet)(MIE_Fq *out, const MIE_Fq *in);

	MIE_Fp *(*FqGetRef)(MIE_Fq *out, size_t idx);
	const MIE_Fp *(*FqGetConstRef)(const MIE_Fq *out, size_t idx);
	void (*FqSetFp)(MIE_Fq *out, size_t idx, const MIE_Fp *in);

	void (*FqAdd)(MIE_Fq *out, const MIE_Fq *in1, const MIE_Fq *in2);
	void (*FqSub)(MIE_Fq *out, const MIE_Fq *in1, const MIE_Fq *in2);
	void (*FqMul)(MIE_Fq *out, const MIE_Fq *in1, const MIE_Fq *in2);

	void (*FqPow)(MIE_Fq *out, const MIE_Fq *in, const unsigned int *tbl, size_t tblSize);

	// functions for Ec(y^2 = x^3 - x + b, b = 1)

	MIE_Ec *(*EcCreate)(void);
	void (*EcDestroy)(MIE_Ec *in);

	MIE_Err (*EcToStr)(char *buf, size_t bufSize, const MIE_Ec *in);

	/**
		set (x, y) of EC by xCandidate
		@param xCandidate [in] candidate of x
		@note x = xCandidate or -xCandidate, LSB of y is always 0 or 1
		@retval 1 success
		@retval 0 failure
	*/
	int (*EcSetX)(MIE_Ec *out, const MIE_Fp *xCandidate);

	void (*EcSet)(MIE_Ec *out, const MIE_Ec *in);

	MIE_Fp *(*EcGetRefX)(MIE_Ec *in);
	MIE_Fp *(*EcGetRefY)(MIE_Ec *in);
	const MIE_Fp *(*EcGetConstRefX)(const MIE_Ec *in);
	const MIE_Fp *(*EcGetConstRefY)(const MIE_Ec *in);

	/**
		(X, Y) sutisfies y^2 = x^3 - x + b
	*/
	int (*EcValid)(const MIE_Ec *in);

	void (*EcAdd)(MIE_Ec *out, const MIE_Ec *in1, const MIE_Ec *in2);

	void (*EcNeg)(MIE_Ec *out, const MIE_Ec *in);

	int (*EcEqual)(const MIE_Ec *in1, const MIE_Ec *in2);

	void (*EcPow)(MIE_Ec *out, const MIE_Ec *in, const unsigned int *tbl, size_t tblSize);

	// calc pairing
	void (*pairing)(MIE_Fq *out, const MIE_Ec *in1, const MIE_Ec *in2);
} MIE_System;

/**
	create factory of Fp/Fq/EC
	@param psys [out] pointer to factory
	@param m [in] degree of field extension(current version supports only m = 193)
*/
MIE_Err MIE_CreateSystem(MIE_System **psys, int m);

/**
	destroy factory
*/
void MIE_DestroySystem(MIE_System *sys);

#ifdef __cplusplus
}
#endif

#endif
