#ifndef MIE_ETAT_SRC_BN_H_
#define MIE_ETAT_SRC_BN_H_
/**
	@file etat_if.h
	@brief BN parameter
	@author herumi
	@note modified new BSD license
	http://www.opensource.org/licenses/bsd-license.php
*/
#include "zm.h"

namespace mie { namespace bn {

static Vuint z, p, r;

typedef mie::ZmZ<mie::Vuint, 1> Fp;

static inline void Init()
{
	z.set("0x6000000000001f2d");
	p = 1 + z * (6 + z * (24 + z * (36 + 36 * z)));
	r = 1 + z * (6 + z * (18 + z * (36 + 36 * z)));
	Fp::setModulo(p);
}

/*
	Fp2 = F[u] / (u^2 + 2)
*/
template<class F>
struct Fp2T : public mie::local::operatable<Fp2T<F> > {
	typedef F Fp;
	F v[2];
	Fp2T()
	{
	}
	Fp2T(int x) { v[0] = x; }
	static inline void add(Fp2T& z, const Fp2T& x, const Fp2T& y)
	{
		z.v[0] = x.v[0] + y.v[0];
		z.v[1] = x.v[1] + y.v[1];
	}
	static inline void sub(Fp2T& z, const Fp2T& x, const Fp2T& y)
	{
		z.v[0] = x.v[0] - y.v[0];
		z.v[1] = x.v[1] - y.v[1];
	}
	/*
		(x0 + x1 u)(y0 + y1 u) = (x0 y0 - 2 x1 y1) + (x0 y1 + x1 y0)u
		                       = (x0 y0 - 2 x1 y1) + ((x0 + x1)(y0 + y1) - x0 y0 - x1 y1)u
	*/
	static inline void mul(Fp2T& z, const Fp2T& x, const Fp2T& y)
	{
		F t0 = x.v[0] * y.v[0];
		F t1 = x.v[1] * y.v[1];
		// don't modify v[0] before calcing v[1] because z may be equal x or y
		z.v[1] = (x.v[0] + x.v[1]) * (y.v[0] + y.v[1]) - t0 - t1;
		z.v[0] = t0 - t1 - t1;
	}
	/*
		(x0 + x1 u)(-u - 1) = (2x1 - x0) + (-x0 - x1)u
	*/
	Fp2T mulByXi() const
	{
		Fp2T z;
		z.v[0] = v[1] + v[1] - v[0];
		z.v[1] = -(v[0] + v[1]);
		return z;
	}
	/*
		(a + bu)^2 = (a - b)(a + 2b) - ab + (2ab)u
	*/
	Fp2T square() const
	{
		Fp2T z;
		F t = v[0] * v[1];
		z.v[0] = (v[0] - v[1]) * (v[0] + v[1] + v[1]) - t;
		z.v[1] = t + t;
		return z;
	}
	/*
		1 / (x0 + x1 u) = (x0 - x1 u) / (x0^2 + 2 x1^2)
	*/
	Fp2T inverse() const
	{
		Fp2T z;
		F t = v[1] * v[1];
		t = v[0] * v[0] + t + t;
		t = t.inverse();
		z.v[0] = v[0] * t;
		z.v[1] = -v[1] * t;
		return z;
	}
	void clear() { v[0].clear(); v[1].clear(); }
	friend std::ostream& operator<<(std::ostream& os, const Fp2T& x)
	{
		return os << x.v[0] << "+" << x.v[1] << "u";
	}
	static inline int compare(const Fp2T& x, const Fp2T& y)
	{
		return x.v[0] == y.v[0] && x.v[1] == y.v[1] ? 0 : 1;
	}
	Fp2T& operator/=(const Fp2T& rhs) { *this *= rhs.inverse(); return *this; }
	inline friend Fp2T operator/(const Fp2T& x, const Fp2T& y) { return x * y.inverse(); }
};

/*
	Fp6T = F[v] / (v^3 - s), s = -u - 1
*/
template<class F>
struct Fp6T : public mie::local::operatable<Fp6T<F> > {
	typedef typename F::Fp Fp;
	F v[3];
	Fp6T()
	{
	}
	Fp6T(const F& x)
	{
		v[0] = x;
	}
	Fp6T(int x)
	{
		v[0].v[0] = x;
	}
	void clear() { v[0].clear(); v[1].clear(); v[2].clear(); }
	static inline void add(Fp6T& z, const Fp6T& x, const Fp6T& y)
	{
		z.v[0] = x.v[0] + y.v[0];
		z.v[1] = x.v[1] + y.v[1];
		z.v[2] = x.v[2] + y.v[2];
	}
	static inline void sub(Fp6T& z, const Fp6T& x, const Fp6T& y)
	{
		z.v[0] = x.v[0] - y.v[0];
		z.v[1] = x.v[1] - y.v[1];
		z.v[2] = x.v[2] - y.v[2];
	}
	static inline void mul(Fp6T& z, const Fp6T& x, const Fp6T& y)
	{
		F t0, t1, t2, t3, t4, t5, t6;
		t0 = x.v[0] * y.v[0];
		t1 = (x.v[1] + x.v[2]) * (y.v[1] + y.v[2]);
		t2 = x.v[1] * y.v[1];
		t3 = x.v[2] * y.v[2];
		t4 = (x.v[0] + x.v[1]) * (y.v[0] + y.v[1]);
		t5 = (x.v[0] + x.v[2]) * (y.v[0] + y.v[2]);
		z.v[0] = t0 + (t1 - t2 - t3).mulByXi();
		z.v[1] = t4 - t0 - t2 + t3.mulByXi();
		z.v[2] = t5 - t0 - t3 + t2;
	}
	static inline int compare(const Fp6T& x, const Fp6T& y)
	{
		return x.v[0] == y.v[0] && x.v[1] == y.v[1] && x.v[2] == y.v[2] ? 0 : 1;
	}
	friend std::ostream& operator<<(std::ostream& os, const Fp6T& x)
	{
		return os << "(" << x.v[0] << ")+(" << x.v[1] << ")v+(" << x.v[2] << ")v^2";
	}
};

typedef Fp2T<mie::bn::Fp> Fp2;
typedef Fp6T<Fp2> Fp6;

} } // mie::bn

#endif
