#ifndef MIE_ETAT_SRC_ETAT_H_
#define MIE_ETAT_SRC_ETAT_H_
/**
	etat pairing on F_3^m
	m : security parameter
*/
#ifndef MIE_DONT_USE_MULTI_CORE
	#define MIE_USE_MULTI_CORE // use two core(only windows)
	#define MIE_USE_MULTI_CORE_WITH_BUSY_LOOP // using busy-loop always for speed-up
#endif
//#define MIE_NEW_VERSION

#define MIE_USE_SSSE3 // use palignr
#define MIE_USE_6OP_ADD

#if defined(MIE_USE_MULTI_CORE) && !defined(_WIN32)
	#undef MIE_USE_MULTI_CORE // only for windows
#endif

#ifdef __INTEL_COMPILER
	#pragma warning(disable : 981)
	#pragma warning(disable : 383)
	#pragma warning(disable : 280) // const
	#pragma warning(disable : 1011) // no return
#endif

#include "etat_util.h"
#include "v128.h"
#include "v256.h"
#ifdef MIE_USE_MULTI_CORE
#include "multi_core.h"
#endif
#include <assert.h>
#include <string>
#if defined(MIE_USE_ETAT_97) || defined(MIE_USE_SSSE3)
	#include "xbyak/xbyak.h"
	#include "xbyak/xbyak_util.h"
#endif

//#define MIE_NOT_PACK_IDX /* a little slow(g_pbit < 4) */

#ifdef MIE_USE_ETAT_97
	#undef MIE_NOT_PACK_IDX
#endif

#ifdef MIE_NOT_PACK_IDX
#define XTBL_SIZE(bit) (1 << (bit * 2))
#else
#define XTBL_SIZE(bit) intpow3T<bit>::val
#endif

namespace MIE_ETAT {

enum {
	g_bit = 3, // shift size in Comb method
	g_pbit = 4, // shift size in Comb methof for mulPara
	g_w = 32 / 4 // byte shift size(32 or 16 or 8)
};

/*
	f(x) = x^m + x^ma + 2
*/
template<int M, int A>
class F3base : private operatable<F3base<M, A> > {
public:
	enum {
		m = M,
		lpN = (m + 1) / 2,
		N = (m + 31) / 32,
		ma = A,
		ALIGN_N = (N + 3) & ~3
	};
//private:
	MIE_ALIGN(16) uint32_t L_[ALIGN_N];
	MIE_ALIGN(16) uint32_t H_[ALIGN_N];
	void set(size_t pos, uint32_t v)
	{
		rawSet(L_, H_, pos, v);
	}
	void setNC(size_t pos, uint32_t v)
	{
		rawSetNC(L_, H_, pos, v);
	}
	uint32_t get(size_t pos) const
	{
		return rawGet(L_, H_, pos);
	}
	uint32_t getL(size_t pos) const
	{
		size_t q = pos / 32;
		size_t r = pos & 31;
		return (L_[q] >> r) & 1;
	}
public:
	F3base() {}
	F3base(const F3base& rhs)
	{
		set(rhs.L_, rhs.H_);
	}
	explicit F3base(const uint32_t *L, const uint32_t *H)
	{
		set(L, H);
	}
	explicit F3base(const char *str)
	{
		setByString(str);
	}
	explicit F3base(uint32_t v)
	{
		set(v);
	}
	void operator=(const F3base& rhs)
	{
		rawCopy(L_, rhs.L_);
		rawCopy(H_, rhs.H_);
	}
	void set(const uint32_t *L, const uint32_t *H)
	{
		rawCopy(L_, L);
		rawCopy(H_, H);
		const int r = m & 31;
		if (r) {
			uint32_t mask = GetMask(r);
			L_[N - 1] &= mask;
			H_[N - 1] &= mask;
		}
	}
	void set(uint32_t v)
	{
		assert(v == 0 || v == 1 || v == 2);
		clear();
		if (v == 1) {
			L_[0] = 1;
		} else if (v == 2) {
			H_[0] = 1;
		}
	}
	void setByString(const char *str)
	{
		clear();
		const size_t len = str ? ::strlen(str) : 0;
		size_t i = 0;
		int pos = static_cast<int>(len) - 1;
		while (i < m && pos >= 0) {
			if (str[pos] != ' ') {
				setNC(i, str[pos] - '0');
				i++;
			}
			pos--;
		}
	}
	void clear()
	{
		rawClear(L_);
		rawClear(H_);
	}
	std::string toString() const
	{
		std::string ret;
		for (int i = m - 1; i >= 0; i--) {
			ret += static_cast<char>(get(i) + '0');
#ifndef MIE_DONT_PRINT_SPACE_FOR_STR
			if (i > 0 && (i & 31) == 0) ret += ' ';
#endif
		}
		return ret;
	}
	MIE_FORCE_INLINE void operator+=(const F3base& b)
	{
		add(*this, *this, b);
	}
	MIE_FORCE_INLINE void operator-=(const F3base& b)
	{
		sub(*this, *this, b);
	}
	MIE_FORCE_INLINE void neg()
	{
		neg(*this, *this);
	}
	MIE_FORCE_INLINE void operator*=(const F3base& rhs)
	{
		mul(*this, *this, rhs);
	}
	MIE_FORCE_INLINE void operator/=(const F3base& rhs)
	{
		F3base x(rhs);
		inverse(x);
		*this *= x;
	}
	static void neg(F3base& out, const F3base& a)
	{
		for (size_t i = 0; i < N; i++) {
			rawNeg(&out.L_[i], &out.H_[i], &a.L_[i], &a.H_[i]);
		}
	}
	static void add(F3base& out, const F3base& a, const F3base& b)
	{
		for (size_t i = 0; i < N; i++) {
			rawAdd(&out.L_[i], &out.H_[i], a.L_[i], a.H_[i], b.L_[i], b.H_[i]);
		}
	}
	static void sub(F3base& out, const F3base& a, const F3base& b)
	{
		for (size_t i = 0; i < N; i++) {
			rawSub(&out.L_[i], &out.H_[i], a.L_[i], a.H_[i], b.L_[i], b.H_[i]);
		}
	}
	/*
		addOut = a + b;
		subOut = a - b;
	*/
	static void addsub(F3base& addOut, F3base& subOut, const F3base& a, const F3base& b)
	{
		F3base t;
		add(t, a, b);
		sub(subOut, a, b);
		addOut = t;
	}
	/*
		add1Out = a1 + b;
		add2Out = a2 + b;
	*/
	static void addPara(F3base& out1, F3base& out2, const F3base& a1, const F3base& a2, const F3base& b)
	{
		F3base t;
		add(t, a1, b);
		add(out2, a2, b);
		out1 = t;
	}

	static MIE_FORCE_INLINE void mul(F3base& out, const F3base& a, const F3base& b)
	{
		F3base xTbl[XTBL_SIZE(g_bit)];
		/*
			total(@Core2Duo 2.66GHz)
					  w
				 32  16   8
			b 2 800 650 575
			  3 712 591 518(= 83 + 434)
			  4 770 623 558(= 230 + 330)
		*/
		makePrevMulTbl<g_bit>(xTbl, a);
		mulLatter<g_bit, g_w>(out, xTbl, b);
	}
	/*
		out1 = a1 * b;
		out2 = a2 * b;
	*/
	static MIE_FORCE_INLINE void mulPara(F3base& out1, F3base& out2, const F3base& a1, const F3base& a2, const F3base& b)
	{
		F3base xTbl[XTBL_SIZE(g_pbit)];
		makePrevMulTbl<g_pbit>(xTbl, b);
		mulLatter<g_pbit, g_w>(out1, xTbl, a1);
		mulLatter<g_pbit, g_w>(out2, xTbl, a2);
	}
#if 0
	template<int bit>
	static void mulLatterC(F3base& out, const F3base *xTbl, const F3base& b)
	{
		const F3base& a = xTbl[1];
		F3base ret, ta(a);
		ret.clear();
		for (int i = 0; i < m; i++) {
			uint32_t t = b.get(i);
			if (t == 1) {
				ret += ta;
			} else if (t == 2) {
				ret -= ta;
			}
			ta.shl1();
		}
		out = ret;
	}
#endif

	typedef void (*mulLatterFunc)(F3base& out, const F3base *xTbl, const F3base& b);
	typedef void (*makePrevMulTblFunc)(F3base *xTbl, const F3base& x);

	template<int bit, int w>
	static void mulLatter(F3base& out, const F3base *xTbl, const F3base& b);

	template<int bit>
	static inline int P(int H, int L)
	{
		return (H << bit) + L;
	}
	template<int bit>
	static void makePrevMulTbl(F3base *xTbl, const F3base& x)
	{
		const bool isRaw = true;
		xTbl[0].clear();
		xTbl[1] = x;
#ifdef MIE_NOT_PACK_IDX
		F3base::neg(xTbl[P<bit>(1, 0)], x); // -x
		F3base y;
		if (bit > 1) {
			F3base t(x); t.shl1(isRaw); // x << 1
			xTbl[P<bit>(0, 2)] = t;
			F3base::neg(xTbl[P<bit>(2, 0)], t);

			F3base::add(y, xTbl[P<bit>(0, 1)], t); // (x << 1) + x
			xTbl[P<bit>(0, 3)] = y;	F3base::neg(xTbl[P<bit>(3, 0)], y);

			F3base::add(y, xTbl[P<bit>(1, 0)], t); // (x << 1) - x
			xTbl[P<bit>(1, 2)] = y; F3base::neg(xTbl[P<bit>(2, 1)], y);
		}
		if (bit > 2) {
			F3base t(xTbl[P<bit>(0, 2)]); t.shl1(isRaw); // x << 2
			xTbl[P<bit>(0, 4)] = t;
			F3base::neg(xTbl[P<bit>(4, 0)], t);

			F3base::add(xTbl[P<bit>(0, 5)], xTbl[P<bit>(0, 1)], t);
			F3base::add(xTbl[P<bit>(1, 4)], xTbl[P<bit>(1, 0)], t);
			F3base::add(xTbl[P<bit>(0, 6)], xTbl[P<bit>(0, 2)], t);
			F3base::add(xTbl[P<bit>(0, 7)], xTbl[P<bit>(0, 3)], t);
			F3base::add(xTbl[P<bit>(1, 6)], xTbl[P<bit>(1, 2)], t);
			F3base::add(xTbl[P<bit>(2, 4)], xTbl[P<bit>(2, 0)], t);
			F3base::add(xTbl[P<bit>(2, 5)], xTbl[P<bit>(2, 1)], t);
			F3base::add(xTbl[P<bit>(3, 4)], xTbl[P<bit>(3, 0)], t);

			F3base::neg(xTbl[P<bit>(4, 1)], xTbl[P<bit>(1, 4)]);
			F3base::neg(xTbl[P<bit>(5, 0)], xTbl[P<bit>(0, 5)]);
			F3base::neg(xTbl[P<bit>(4, 2)], xTbl[P<bit>(2, 4)]);
			F3base::neg(xTbl[P<bit>(4, 3)], xTbl[P<bit>(3, 4)]);
			F3base::neg(xTbl[P<bit>(5, 2)], xTbl[P<bit>(2, 5)]);
			F3base::neg(xTbl[P<bit>(6, 0)], xTbl[P<bit>(0, 6)]);
			F3base::neg(xTbl[P<bit>(6, 1)], xTbl[P<bit>(1, 6)]);
			F3base::neg(xTbl[P<bit>(7, 0)], xTbl[P<bit>(0, 7)]);
		}
#else
		F3base::neg(xTbl[2], xTbl[1]);
		if (bit > 1) {
			xTbl[3] = xTbl[1]; xTbl[3].shl1(isRaw);
			F3base::add(xTbl[4], xTbl[1], xTbl[3]);
			F3base::add(xTbl[5], xTbl[2], xTbl[3]);
			F3base::neg(xTbl[6], xTbl[3]);
			F3base::neg(xTbl[7], xTbl[5]);
			F3base::neg(xTbl[8], xTbl[4]);
		}
		if (bit > 2) {
			const int t = intpow3T<2>::val;
			xTbl[t] = xTbl[t / 3]; xTbl[t].shl1(isRaw);
			for (int i = 0; i < t - 1; i++) {
				F3base::add(xTbl[i + t + 1], xTbl[i + 1], xTbl[t]);
			}
			static const int revTbl[] = { 9, 11, 10, 15, 17, 16, 12, 14, 13 };
			for (size_t i = 0; i < mie::NumOfArray(revTbl); i++) {
				F3base::neg(xTbl[i + t * 2], xTbl[revTbl[i]]);
			}
		}
		if (bit > 3) {
			const int t = intpow3T<3>::val;
			xTbl[t] = xTbl[t / 3]; xTbl[t].shl1(isRaw);
			for (int i = 0; i < t - 1; i++) {
				F3base::add(xTbl[i + t + 1], xTbl[i + 1], xTbl[t]);
			}
			static const int revTbl[] = {
				27, 29, 28, 33, 35, 34, 30, 32, 31,
				45, 47, 46, 51, 53, 52, 48, 50, 49,
				36, 38, 37, 42, 44, 43, 39, 41, 40,
			};
			for (size_t i = 0; i < mie::NumOfArray(revTbl); i++) {
				F3base::neg(xTbl[i + t * 2], xTbl[revTbl[i]]);
			}
		}
#endif
	}

	/* return x^  3 */
	typedef void (*cubeFunc)(F3base& x);
	static void cube(F3base& x);
#if 0
	{
		uint32_t L[N * 3];
		uint32_t H[N * 3];
		rawClear(L);
		rawClear(H);
		for (int i = 0; i < m; i++) {
			rawSet(L, H, i * 3, x.get(i));
		}
		for (int i = m * 3 - 1; i >= m; i--) {
			uint32_t t, x = rawGet(L, H, i);
			t = (rawGet(L, H, i - (m - ma)) - x + 3) % 3;
			rawSet(L, H, i - (m - ma), t);
			t = (rawGet(L, H, i - m) + x) % 3;
			rawSet(L, H, i - m, t);
			rawSet(L, H, i, 0);
		}
		x = F3base(L, H);
	}
#endif

	typedef void (*cubeSeqFunc)(F3base *);
	static MIE_FORCE_INLINE void cubeSeq(F3base *tbl)
	{
		F3base t = tbl[0];
		for (size_t i = 1; i < lpN; i++) {
			cube(t);
			tbl[i] = t;
		}
	}

	typedef void (*rawShlFunc)(F3base& x, int bit);
	static void rawShl(F3base& x, int bit)
	{
		assert(bit < m);
		int n = bit >> 5;
		if (n > 0) {
			int i = N - 1;
			for (; i >= n; i--) {
				x.L_[i] = x.L_[i - n];
				x.H_[i] = x.H_[i - n];
			}
			for (; i >= 0; i--) {
				x.L_[i] = 0;
				x.H_[i] = 0;
			}
			bit &= 31;
		}
		if (bit == 0) return;
		uint32_t L = 0, H = 0;
		for (size_t i = 0; i < N; i++) {
			uint32_t t;
			t = x.L_[i];
			x.L_[i] = (t << bit) | L;
			L = t >> (32 - bit);
			t = x.H_[i];
			x.H_[i] = (t << bit) | H;
			H = t >> (32 - bit);
		}
	}

	typedef int (*degreeFunc)(const F3base& x);
	static int degree(const F3base& x)
	{
		for (int i = m - 1; i >= 0; i--) {
			if (x.get(i)) return i;
		}
		return 0;
	}

	static const uint32_t powTblForSquareRoot[];
	static const uint32_t powTblSize;
	static bool squareRoot(F3base& x)
	{
		F3base t(x);
		power(t, powTblForSquareRoot, powTblSize);
		if (t * t == x) {
			x = t;
			return true;
		} else {
			return false;
		}
	}

	bool isZero() const
	{
		for (size_t i = 0; i < N; i++) {
			if (L_[i] | H_[i]) return false;
		}
		return true;
	}
	bool operator==(const F3base& rhs) const
	{
		for (size_t i = 0; i < N; i++) {
			if ((L_[i] ^ rhs.L_[i]) | (H_[i] ^ rhs.H_[i])) return false;
		}
		return true;
	}
	bool operator!=(const F3base& rhs) const { return !(*this == rhs); }
	void shl1(bool isRaw = false)
	{
		rawShl(*this, 1);
		if (!isRaw) {
			uint32_t t = get(m);
			if (t) {
				L_[N - 1] &= GetMask(m & 31);
				H_[N - 1] &= GetMask(m & 31);
				reduction1(t);
			}
		}
	}
	void shl(int bit, bool isRaw = false)
	{
		if (isRaw) {
			rawShl(*this, bit);
		} else {
			for (int i = 0; i < bit; i++) shl1();
		}
	}

	static MIE_FORCE_INLINE void cubeN(F3base& x, int n)
	{
		for (int i = 0; i < n; i++) {
			cube(x);
		}
	}
	void reduction1(uint32_t t)
	{
		uint32_t x;
		x = (get(ma) + t * 2) % 3;
		set(ma, x);
		x = (get(0) + t) % 3;
		set(0, x);
	}
	void swap(F3base& x)
	{
		for (size_t i = 0; i < N; i++) {
			std::swap(L_[i], x.L_[i]);
			std::swap(H_[i], x.H_[i]);
		}
	}
};

template<class F>
static void inverse(F& x)
{
	static const F one(1);
	static const F two(2);
	if (x == one || x == two) return;
	static bool init = false;
	static F mod("0");
	if (!init) {
		init = true;
		mod.setNC(F::ma, 1);
		mod.setNC(0, 2);
	}
	F u(x), v(x), g1, g2(1);
	int ud = F::degree(u);
	int j = F::m - ud;
	int c = u.get(ud);
	if (c == 1) u.neg(); /* -(u_deg(u))u; set MSB bit = 2 */
	u.set(ud, 0);
	F::rawShl(u, j);
	u += mod; /* f - (u_deg(u)) u x^j; add f */
	g1.clear(); g1.set(j, 3 - c);/* g1 = -u_deg(u) x^j */
	for (;;) {
		ud = F::degree(u);
		if (ud == 0) break;
		int vd = F::degree(v);
		j = ud - vd;
		if (j < 0) {
			v.swap(u);
			g1.swap(g2);
			j = -j;
			std::swap(ud, vd);
		}
		/*
			always u.get(ud) != 0 and v.get(vd) != 0
			c == 1 iff uL ^ vL == 0
		*/
		c = !(u.getL(ud) ^ v.getL(vd));
		F t(v);
		if (c) {
			t.neg();
		}
		F::rawShl(t, j);
		u += t;
		t = g2;
		if (c) {
			t.neg();
		}
		F::rawShl(t, j);
		g1 += t;
	}
	if (u.H_[0]) g1.neg();
	x = g1;
}

#ifndef MIE_USE_ETAT_97

#if 0 /* for general */
/**
	[a:b] -> 3a + b
	[a:b:c] -> 9a + 3b + c
	[a:b:c:d] -> 27a + 9b + 3c + d
*/
template<int bit, int w>
static inline uint32_t getTri(uint32_t x, int i)
{
	uint32_t mask = (w == 8) ? 0x01010101 : (w == 16) ? 0x00010001 : (w == 32) & 0x00000001;
	uint32_t v = 0;
	for (int j = 0; j < bit; j++) {
		v += (x >> (i + bit - 1 - j)) & mask;
		if (j < bit - 1) {
			v *= 3;
		}
	}
	return v;
}

template<int bit, int w, class T>
static inline void setIdx(uint8_t *p, const T& x, int i)
{
	for (int j = 0; j < T::N - 1; j++) {
		uint32_t L = getTri<bit, w>(x.L_[j], i);
		uint32_t H = getTri<bit, w>(x.H_[j], i);
		*((uint32_t*)p + j) = L + H * 2;
	}
}

template<int bit, int w, class T>
void makeIdxTbl(uint8_t idxTbl[(w + bit - 1) / bit][((T::N - 1 + 3) & ~3) * 4], const T& b)
{
	const int topBit = ((w - 1) / bit) * bit;
	const int firstShiftBit = w - topBit < bit ? w - topBit : bit;

	setIdx<firstShiftBit, w>(&idxTbl[0][0], b, topBit);
	for (int s = 1, i = topBit - bit; i >= 0; i -= bit, s++) {
		setIdx<bit, w>(&idxTbl[s][0], b, i);
	}
}
#endif

/*
	2m bit size temporary variable for mul
*/
template<class F>
class F3baseDbl {
public:
	enum {
		m = F::m,
		N = F::N,
		ma = F::ma,
		m2 = F::m * 2,
		N2 = (m2 + 31) / 32,
		ALIGN_N2 = (N2 + 3) & ~3
	};
private:
	MIE_ALIGN(16) uint32_t L_[ALIGN_N2];
	MIE_ALIGN(16) uint32_t H_[ALIGN_N2];
	void clear()
	{
		rawClear(L_);
		rawClear(H_);
	}
public:
	F3baseDbl()
	{
	}
	F3baseDbl(const F& rhs)
	{
		rawCopy(L_, rhs.L_, N);
		rawCopy(H_, rhs.H_, N);
		::memset(&L_[N], 0, sizeof(*L_) * (N2 - N));
		::memset(&H_[N], 0, sizeof(*H_) * (N2 - N));
	}
	F3baseDbl(const uint32_t *L, const uint32_t *H)
	{
		rawCopy(L_, L, N2);
		rawCopy(H_, H, N2);
	}
	void store(uint32_t (&L)[N2], uint32_t (&H)[N2]) const
	{
		rawCopy(L, L_, N2);
		rawCopy(H, H_, N2);
	}
	uint32_t *getL() { return L_; }
	uint32_t *getH() { return H_; }
	template<int bit>
	void setTop(const F& rhs)
	{
		const int q = bit / 32;
		::memset(L_, 0, q * sizeof(uint32_t));
		::memset(H_, 0, q * sizeof(uint32_t));
		::memcpy(&L_[q], rhs.L_, N * sizeof(uint32_t));
		::memcpy(&H_[q], rhs.H_, N * sizeof(uint32_t));
		::memset(&L_[q] + N, 0, (ALIGN_N2 - N - q) * sizeof(uint32_t));
		::memset(&H_[q] + N, 0, (ALIGN_N2 - N - q) * sizeof(uint32_t));
		const int r = bit & 31;
		if (r) {
			shl<r>();
		}
	}
	void add(const F& rhs)
	{
		for (int i = 0; i < N; i++) {
			rawAdd((uint32_t*)((uint8_t*)&L_[i]), (uint32_t*)((uint8_t*)&H_[i]), rhs.L_[i], rhs.H_[i]);
		}
	}
	/*
		byte rotr
	*/
	template<int nByte>
	void rotrByte()
	{
		if (nByte == 0) return;
		const int range = N2 & ~3;
		const int n = nByte * 8;
		const int q = n / 32;
		const int r = n & 31;
		if (q > 0) {
			uint32_t cL_[N2];
			uint32_t cH_[N2];
			rawCopy(cL_, L_);
			rawCopy(cH_, H_);
			for (int i = 0; i < range; i++) {
				L_[i] = cL_[(i + q) % range];
				H_[i] = cH_[(i + q) % range];
			}
		}
		if (r > 0) {
			uint32_t L = L_[0];
			uint32_t H = H_[0];
			for (int i = 0; i < range - 1; i++) {
				L_[i] = (L_[i] >> r) | (L_[i + 1] << (32 - r));
				H_[i] = (H_[i] >> r) | (H_[i + 1] << (32 - r));
			}
			L_[range - 1] = (L_[range - 1] >> r) | (L << (32 - r));
			H_[range - 1] = (H_[range - 1] >> r) | (H << (32 - r));
		}
	}
	/*
		bit rotl (nByte must be multiplier of 4)
	*/
	template<int nByte>
	void rotlByte()
	{
		if (nByte == 0) return;
		const int range = N2 & ~3;
		const int n = nByte * 8;
		const int q = n / 32;
		if (q > 0) {
			uint32_t cL_[N2];
			uint32_t cH_[N2];
			rawCopy(cL_, L_);
			rawCopy(cH_, H_);
			for (int i = 0; i < range; i++) {
				L_[i] = cL_[(i - q + range) % range];
				H_[i] = cH_[(i - q + range) % range];
			}
		}
		assert((n & 31) == 0);
#if 0 /* avoid compile error for r == 0 */
		const int r = n & 31;
		if (r > 0) {
			uint32_t L = L_[range - 1];
			uint32_t H = H_[range - 1];
			for (int i = range - 1; i > 0; i--) {
				L_[i] = (L_[i] << r) | (L_[i - 1] >> (32 - r));
				H_[i] = (H_[i] << r) | (H_[i - 1] >> (32 - r));
			}
			L_[0] = (L_[0] << r) | (L >> (32 - r));
			H_[0] = (H_[0] << r) | (H >> (32 - r));
		}
#endif
	}
	static inline void put(const char *msg, const uint32_t *p)
	{
		printf("%s ", msg);
		for (int i = 0; i < N2; i++) {
			printf("%08x:", p[N2 - i - 1]);
		}
		printf("\n");
	}
	/*
		srcL, srcH >= 2m bit
		MSB of data will be vanished
	*/
	template<int bit>
	void shl()
	{
		assert(0 < bit && bit < 32);
		uint32_t t;
		t = 0;
		for (size_t i = 0; i < N2; i++) {
			uint32_t x = L_[i];
			L_[i] = (x << bit) | t;
			t = x >> (32 - bit);
		}
		t = 0;
		for (size_t i = 0; i < N2; i++) {
			uint32_t x = H_[i];
			H_[i] = (x << bit) | t;
			t = x >> (32 - bit);
		}
	}
	bool operator==(const F3baseDbl& rhs) const
	{
		for (int i = 0; i < N2; i++) {
			if (L_[i] != rhs.L_[i] || H_[i] != rhs.H_[i]) return false;
		}
		return true;
	}
	bool operator!=(const F3baseDbl& rhs) const
	{
		return !(*this == rhs);
	}
	std::string toHex(const uint32_t *x, int n) const
	{
		std::string ret;
		for (int i = n - 1; i >= 0; i--) {
			char buf[32];
			sprintf(buf, "%08x", x[i]);
			ret += buf;
			if (i > 0) ret += ':';
		}
		return ret;
	}
	std::string toString() const
	{
		return std::string("\tL ") + toHex(L_, N2) + "\n\tH " + toHex(H_, N2);
	}
	void reduce(F& ret)
	{
		for (int i = m * 2 - 1; i >= m; i--) {
			uint32_t t, x = rawGet(L_, H_, i);
			t = (rawGet(L_, H_, i - ((int)m - ma)) - x + 3) % 3;
			rawSet(L_, H_, i - ((int)m - ma), t);
			t = (rawGet(L_, H_, i - m) + x) % 3;
			rawSet(L_, H_, i - m, t);
			rawSet(L_, H_, i, 0);
		}
		ret = F(L_, H_);
	}
};
#endif

/*
	specialization of F3base
*/
#ifdef MIE_USE_ETAT_97
	#include "f3base97.h"
#endif

#ifdef MIE_USE_ETAT_193
	#include "f3base193.h"
#endif

#include "extended_f.h"

/*
	specialization of extended F3base
*/
#ifdef MIE_USE_ETAT_97
	typedef MIE_ETAT::ECgeneral<F3_97> EConF3_97;
	typedef MIE_ETAT::F3_3base<F3_97> F3_97_3;
	typedef MIE_ETAT::F6base<F3_97> F3_97_6;
#endif
#ifdef MIE_USE_ETAT_193
	typedef MIE_ETAT::ECgeneral<F3_193> EConF3_193;
	typedef MIE_ETAT::F3_3base<F3_193> F3_193_3;
	typedef MIE_ETAT::F6base<F3_193> F3_193_6;
#endif

static bool init()
{
#if defined(MIE_USE_ETAT_97) || defined(MIE_USE_SSSE3)
	Xbyak::util::Cpu cpu;
	if (!cpu.has(Xbyak::util::Cpu::tSSSE3)) {
		fprintf(stderr, "This program can't run because this cpu doesn't support SSSE3.\n");
		return false;
	}
#endif
#ifdef MIE_USE_MULTI_CORE
	s_parallelizer.init();
#endif
	return true;
}

} // end of namespace

#endif
