#ifndef MIE_CLOCK_H_
#define MIE_CLOCK_H_
/**
	@file mie_cpuclock.h
	@author $Author: shigeo $
	@version $Revision: 1.12 $
	@date $Date: 2010/04/26 05:37:29 $
*/
#include "mie_common.h"
#ifdef _WIN32
	#include <windows.h>
	#include <intrin.h>
	#include <ctime>
#else
	#include <sys/time.h>
	#include <cstring>
	#include <cstdlib>
#endif

namespace mie {

/**
	CPU clock class
*/
class Clock {
	uint64_t getRdtsc() const
	{
#ifdef _WIN32
		return __rdtsc();
#else
		unsigned int eax, edx;
		__asm__ volatile("rdtsc" : "=a"(eax), "=d"(edx));
		return ((uint64_t)edx << 32) | eax;
#endif
	}
public:
	Clock()
		: cpuFreqHz_(1)
		, clock_(0)
		, count_(0)
	{
	}

	/**
		get freq of cpu
	*/
	void detectCpuFreqHz()
	{
#ifdef _WIN32
		LARGE_INTEGER freq;
		QueryPerformanceFrequency(&freq);
		cpuFreqHz_ = freq.QuadPart;
#else
		FILE *fp = fopen("/proc/cpuinfo", "r");
		if (fp) {
			char buf[128];
			while (fgets(buf, sizeof(buf), fp)) {
				if (memcmp(buf, "cpu MHz", 7) == 0) {
					const char *p = std::strchr(buf, ':');
					if (p) {
						cpuFreqHz_ = uint64_t(std::atof(p + 2) * 1e6);
						break;
					}
				}
			}
			fclose(fp);
		}
#endif
	}

	double getCpuFreqHz() const { return static_cast<double>(cpuFreqHz_); }

	static inline double getCurrentTimeSec()
	{
#ifdef _WIN32
		return std::clock() / double(CLOCKS_PER_SEC);
#else
		struct timeval tv;
		gettimeofday(&tv, NULL);
		return tv.tv_sec + (double) tv.tv_usec * 1e-6;
#endif
	}

	/**
		relative passed time(nsec)
	*/
	inline double getRelativeNsec() const
	{
		return getRdtsc() * 1.0e9 / cpuFreqHz_;
	}
	void begin()
	{
		clock_ -= getRdtsc();
	}
	void end()
	{
		clock_ += getRdtsc();
		count_++;
	}
	int getCount() const { return count_; }
	uint64_t getClock() const { return clock_; }
	void put()
	{
		if (count_) {
			printf("call %dtimes:ave %.3fKclk\n", count_, double(clock_) / count_ / 1000);
		}
	}
private:
	uint64_t cpuFreqHz_;
	uint64_t clock_;
	int count_;
};

} // end of namespace

#endif /* MIE_CLOCK_H_ */
