#ifndef MIE_SOCKET_H_
#define MIE_SOCKET_H_
/**
	@file mie_socket.h
	@brief tiny socket class
	@author $Author: shigeo $
*/
#include <cassert>
#include <errno.h>
#ifdef _WIN32
	#include <winsock2.h>
	#include <ws2tcpip.h> // for socklen_t
	#if _MSC_VER > 1200
		#pragma warning(push)
		#pragma warning(disable : 4996) /* sscanf */
		#pragma warning(disable : 4127) // constant if
	#endif
	#pragma comment(lib, "ws2_32.lib")
	#pragma comment(lib, "iphlpapi.lib")
#else
	#include <unistd.h>
	#include <sys/socket.h>
	#include <netinet/tcp.h>
	#include <netdb.h>
	#include <memory.h>
	#include <signal.h>
#endif
#include "mie_common.h"

namespace mie { namespace socket_local {

#ifdef _WIN32
	typedef SOCKET SocketHandle;
#else
	typedef int SocketHandle;
#endif

	static inline ssize_t write(SocketHandle sd, const char *buf, size_t len)
	{
#ifdef _WIN32
		return ::send(sd, buf, (int)len, 0);
#else
		ssize_t ret;
		do {
			ret = ::write(sd, buf, len);
		} while (ret < 0 && errno == EINTR);
		return ret;
#endif
	}
	static inline ssize_t read(SocketHandle sd, char *buf, size_t len)
	{
#ifdef _WIN32
		return ::recv(sd, buf, (int)len, 0);
#else
		ssize_t ret;
		do {
			ret = ::read(sd, buf, len);
		} while (ret < 0 && errno == EINTR);
		return ret;
#endif
	}
	static inline void close(SocketHandle sd)
	{
#ifdef _WIN32
		::shutdown(sd, SD_SEND);
		::closesocket(sd);
#else
		::close(sd);
#endif
	}
} }

//#define MIE_SOCKET_DEBUG

#ifdef NDEBUG
	#undef MIE_SOCKET_DEBUG
#endif

#if defined(MIE_SOCKET_DEBUG) && !defined(MIE_SOCKET_DPRINTF)
	#include <stdio.h>
	#define MIE_SOCKET_DPRINTF(x) printf x
#else
	#define MIE_SOCKET_DPRINTF(x)
#endif

namespace mie {

class SocketAddr {
	struct sockaddr_in addr_;
public:
	SocketAddr()
	{
	}
	SocketAddr(const char *address, int port)
	{
		set(address, port);
	}
	bool setName(const char *address)
	{
		memset(&addr_, 0, sizeof(addr_));
		addr_.sin_family = AF_INET;

#ifdef _WIN32
		struct hostent *host = gethostbyname(address); /* return value is thread local */
		if (!host) {
			return false;
		}
		memcpy(&addr_.sin_addr, host->h_addr_list[0], host->h_length);
#else
		{
			int err;
			char buf[128];
			struct hostent host, *p;
			if (gethostbyname_r(address, &host, buf, sizeof(buf), &p, &err)) {
				return false;
			}
			memcpy(&addr_.sin_addr, host.h_addr_list[0], host.h_length);
		}
#endif
		return true;
	}
	void setPort(int port)
	{
		addr_.sin_port = htons((unsigned short)port);
	}
	void set(const char *address, int port)
	{
		setName(address);
		setPort(port);
	}
	const struct sockaddr *get() const { return (const struct sockaddr*)&addr_; }
};
/*
	\PbgNX
	@note Rs[ƏLړ
*/
class Socket {
private:
	enum {
		isTCP = true
	};
	socket_local::SocketHandle sd_;
#ifdef WIN32
	bool setTimeout(int type, int msec)
	{
		return setSocketOption(type, msec);
	}
	int getTimeout(int type) const
	{
		int val;
		if (getSocketOption(type, &val)) {
			return val;
		} else {
			return -1;
		}
	}
#else
	bool setTimeout(int type, int msec)
	{
		struct timeval t;
		t.tv_sec = msec / 1000;
		t.tv_usec = (msec % 1000) * 1000;
		return setsockopt(sd_, SOL_SOCKET, type, reinterpret_cast<const char*>(&t), sizeof(t)) == 0;
	}
	int getTimeout(int type) const
	{
		struct timeval t;
		socklen_t len = sizeof(t);
		if (getsockopt(sd_, SOL_SOCKET, type, reinterpret_cast<char*>(&t), &len) == 0) {
			return t.tv_sec * 1000 + t.tv_usec / 1000; /* msec */
		} else {
			return -1;
		}
	}
#endif
	int getError() const
	{
#ifdef _WIN32
		return (WSAGetLastError() == WSAETIMEDOUT) ? TIMEOUT : ERR;
#else
		return (errno == EAGAIN || errno == EWOULDBLOCK) ? TIMEOUT : ERR;
#endif
	}
public:
	enum {
		NOERR = 1,
		CLOSED = 0,
		TIMEOUT = -1,
		ERR = -2
	};
#ifndef _WIN32
	enum {
		INVALID_SOCKET = -1
	};
#endif

	/** \PbgVXe(xĂяo΂悢) */
	static inline void InitSystem()
	{
#ifdef _WIN32
		WSADATA data;
		WSAStartup(MAKEWORD(2, 2), &data);
#else
		::signal(SIGPIPE, SIG_IGN);
#endif
	}

	/** \PbgVXeI */
	static inline void TermSystem()
	{
#ifdef _WIN32
		WSACleanup();
#endif
	}

	Socket()
		: sd_(INVALID_SOCKET)
	{
	}

	bool isValid() const { return sd_ != INVALID_SOCKET; }

	Socket(Socket& rhs)
		: sd_(rhs.sd_)
	{
		rhs.sd_ = INVALID_SOCKET;
	}
	void operator=(Socket& rhs)
	{
		close();
		sd_ = rhs.sd_; rhs.sd_ = INVALID_SOCKET;
//		sd_ = mie::AtomicExchange((uint32_t*)&rhs.sd_, INVALID_SOCKET);
	}

	~Socket()
	{
		close();
	}

	bool close()
	{
		if (!isValid()) return true;

		mie::socket_local::close(sd_);
		sd_ = INVALID_SOCKET;
		return true;
	}

	/*!
		f[^𑗐M
		@param buf [out] Mobt@
		@param bufSize [in] Mobt@TCY(byte)
		@retval >= 0 MTCY
		@retval TIMEOUT ^CAEg
		@retval CLOSED \Pbgꂽ
		@retval ERR ̑̃G[
	*/
	ssize_t write(const char *buf, size_t bufSize)
	{
		int ret;
		ret = mie::socket_local::write(sd_, buf, bufSize);
		if (ret >= 0) return ret; /* bufSize == 0Ȃ0oCgf[^̃pPbg𑗐M */
		return getError();
	}

	/*!
		f[^M
		@param buf [out] Mobt@
		@param bufSize [in] Mobt@TCY(byte)
		@retval > 0 I(MTCY)
		@retval TIMEOUT ^CAEg
		@retval CLOSED \Pbgꂽ
		@retval ERR ̑̃G[
	*/
	ssize_t read(char *buf, size_t bufSize)
	{
		int ret;
		ret = mie::socket_local::read(sd_, buf, bufSize);
		if (ret >= 0) return ret;
		return getError();
	}

	/*!
		^CAEgȂw肳ꂽTCYM܂read()D
		@param buf [out] Mobt@
		@param bufSize [in] Mobt@TCY(byte)
		@return NOERR I
		@retval TIMEOUT ^CAEg
		@retval CLOSED \Pbgꂽ
		@retval ERR ̑̃G[
	*/
	int readAll(char *buf, size_t bufSize)
	{
		while (bufSize > 0) {
			ssize_t ret = read(buf, bufSize);
			if (ret <= 0) return ret;
			buf += ret;
			bufSize -= ret;
		}
		return NOERR;
	}
	/*!
		^CAEgȂw肳ꂽTCY𑗐MD
		@param buf [out] Mobt@
		@param bufSize [in] Mobt@TCY(byte)
		@return NOERR I
		@retval TIMEOUT ^CAEg
		@retval CLOSED \Pbgꂽ
		@retval ERR ̑̃G[
	*/
	int writeAll(const char *buf, size_t bufSize)
	{
		while (bufSize > 0) {
			ssize_t ret = write(buf, bufSize);
			if (ret <= 0) return ret;
			buf += ret;
			bufSize -= ret;
		}
		return NOERR;
	}
	/**
		NCAgƂă|[g
		@param address [in] AhX
		@param port [in] |[gԍ
	*/
	bool connect(const char *address, int port)
	{
		SocketAddr addr;
		addr.setName(address);
		addr.setPort(port);
		return connect(addr);
	}
	/**
		Omie::SocketAddrŐڑ
	*/
	bool connect(const mie::SocketAddr& addr)
	{
		sd_ = socket(AF_INET, isTCP ? SOCK_STREAM : SOCK_DGRAM, 0);
		if (!isValid()) {
			return false;
		}

		if (::connect(sd_, addr.get(), sizeof(addr)) == 0) {
			return true;
		}
		int keep = errno;
		close();
		errno = keep;
		return false;
	}

	/**
		(T[op)
		@param port [in] |[gԍ
	*/
	bool bind(int port)
	{
		sd_ = ::socket(AF_INET, isTCP ? SOCK_STREAM : SOCK_DGRAM, IPPROTO_TCP);
		if (!isValid()) {
			return false;
		}
		if (!setSocketOption(SO_REUSEADDR, 1)) {
			return false;
		}
		struct sockaddr_in addr;
		memset(&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_port = htons((unsigned short)port);
		addr.sin_addr.s_addr = htonl(INADDR_ANY);

		if (::bind(sd_, (struct sockaddr *)&addr, sizeof(addr)) == 0) {
			if (!isTCP || ::listen(sd_, SOMAXCONN) == 0) {
				return true;
			}
		}
		mie::socket_local::close(sd_);
		sd_ = INVALID_SOCKET;
		return false;
	}

	/*!
		acceptłȂtrue, ʖڂȂfalse
		1bԑ҂đʖڂȂfalseŔ
		while (!server.queryAccept()) {
		}
		client.accept(server);
	*/
	bool queryAccept()
	{
		assert(isTCP);
		struct timeval timeout;
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
		fd_set fds;
		FD_ZERO(&fds);
		FD_SET((unsigned)sd_, &fds);
		int ret = ::select((int)sd_ + 1, &fds, 0, 0, &timeout);
		if (ret > 0) {// && ret != SOCKET_ERROR) {
			return FD_ISSET(sd_, &fds) != 0;
		}
		return false;
	}

	/**
		҂󂯏(T[op)
	*/
	bool accept(Socket& client, struct in_addr *ip = 0) const
	{
		assert(isTCP);
		struct sockaddr_in addr;
		socklen_t len = sizeof(addr);
		client.sd_ = ::accept(sd_, (struct sockaddr *)&addr, &len);
		if (client.isValid()) {
			if (ip) *ip = addr.sin_addr;
			return true;
		}
		return false;
	}

	/**
		intpsetsockopt()
	*/
	bool setSocketOption(int level, int optname, int value)
	{
		return setsockopt(sd_, level, optname, reinterpret_cast<const char*>(&value), sizeof(value)) == 0;
	}
	bool setSocketOption(int optname, int value)
	{
		return setSocketOption(SOL_SOCKET, optname, value);
	}
	/**
		intpgetsockopt()
	*/
	bool getSocketOption(int level, int optname, int* value) const
	{
		socklen_t len = sizeof(*value);
		return getsockopt(sd_, level, optname, reinterpret_cast<char*>(value), &len) == 0;
	}
	bool getSocketOption(int optname, int* value) const
	{
		return getSocketOption(SOL_SOCKET, optname, value);
	}
	template<typename T>
	bool setSocketOption(int optname, const T& value)
	{
		return setsockopt(sd_, SOL_SOCKET, optname, reinterpret_cast<const char*>(&value), sizeof(T)) == 0;
	}
	template<typename T>
	bool getSocketOption(int optname, T* value) const
	{
		socklen_t len = sizeof(T);
		return getsockopt(sd_, SOL_SOCKET, optname, reinterpret_cast<char*>(value), &len) == 0;
	}
	/**
		lingerݒ
	*/
	bool setLinger(unsigned short l_onoff, unsigned short l_linger)
	{
		struct linger linger;
		linger.l_onoff = l_onoff;
		linger.l_linger = l_linger;
		return setSocketOption(SO_LINGER, &linger);
	}
	/**
		Mobt@TCY̎擾
		@retval  obt@TCY(byte)
		@retval -1 G[
	*/
	int getReceiveBufferSize() const
	{
		int val;
		if (getSocketOption(SO_RCVBUF, &val)) {
			return val;
		} else {
			return -1;
		}
	}
	/**
		Mobt@̐ݒ
		@param size [in] obt@TCY(byte)
	*/
	bool setReceiveBufferSize(int size)
	{
		return setSocketOption(SO_RCVBUF, size);
	}
	/**
		Mobt@TCY̎擾
		@retval  obt@TCY(byte)
		@retval -1 G[
	*/
	int getSendBufferSize() const
	{
		int val;
		if (getSocketOption(SO_SNDBUF, &val)) {
			return val;
		} else {
			return -1;
		}
	}
	/**
		Mobt@̐ݒ
		@param size [in] obt@TCY(byte)
	*/
	bool setSendBufferSize(int size)
	{
		return setSocketOption(SO_SNDBUF, size);
	}
public:
	/**
		M^CAEg̐ݒ
		@param msec [in] ~b
	*/
	bool setSendTimeout(int msec)
	{
		return setTimeout(SO_SNDTIMEO, msec);
	}
	/**
		M^CAEg̐ݒ
		@param msec [in] ~b
	*/
	bool setReceiveTimeout(int msec)
	{
		return setTimeout(SO_RCVTIMEO, msec);
	}
	/**
		M^CAEg̎擾
	*/
	int getSendTimeout() const
	{
		return getTimeout(SO_SNDTIMEO);
	}
	/**
		M^CAEg̎擾
	*/
	int getReceiveTimeout() const
	{
		return getTimeout(SO_RCVTIMEO);
	}
};

} // end of namespace

#if _MSC_VER > 1200
	#pragma warning(pop)
#endif

#endif /* MIE_SOCKET_H_ */
